import os

from langfuse.callback import CallbackHandler


# langfuse_handler = CallbackHandler(
#     public_key=os.environ["LANGFUSE_PUBLIC_KEY"],
#     secret_key=os.environ["LANGFUSE_SECRET_KEY"],
#     host=os.environ["LANGFUSE_HOST"],
# )